//
//  otpCE - One-Time Password Calculater for Windows CE
//
//  Copyright(C) 1997,1998 NEC Corporation
//	All rights reserved.
//	Author: Mizukoshi, Yasuhiro
//	E-Mail: pochi@ccs.mt.nec.co.jp
//

static char CopyRightID[] = "Copyright(C) 1997,1998 NEC Corporation. All rights reserved.";

#include    <windows.h>
#include	<winuser.h>
#include	<stdlib.h>

#define		GLOBALS
#include	"otpce.h"

//
//	InitOtpCE		RegisterClass, LoadIcon
//	CheckOption		Check "/dlg"
//	WinMain
//	WndProc
//	CalcOTP			Get strings from the clipboard and check it out.
//	PassphraseDlgProc
//	FindSeq
//	ScanBuf
//	SetFuncName
//	DialogDlgProc	is used when dialog mode is set.
//

int InitOtpCE()
{
	WNDCLASS	wc;

	szAppName	= OTPCE_NAME;

	hOTPIcon	= LoadIcon(hInst, OTPCE_ICON);

	bSKeyMD5	= FALSE;
	LoadRegistory();

	wc.style		  = 0;
	wc.lpfnWndProc	  = (WNDPROC) WndProc;
	wc.cbClsExtra	  = 0;
	wc.cbWndExtra	  = 0;
	wc.hInstance	  = hInst;
	wc.hIcon		  = NULL;
	wc.hCursor		  = NULL;
	wc.hbrBackground  = (HBRUSH) GetStockObject(WHITE_BRUSH);
	wc.lpszMenuName	  = NULL;
	wc.lpszClassName  = szAppName;

	RegisterClass(&wc);

	return TRUE;
}

BOOL LoadRegistory()
{
	HKEY  hKey;
	DWORD dwType;
	DWORD dwLength;
	DWORD dwData;

	if (RegOpenKeyEx(HKEY_CURRENT_USER, OTPCE_SUBKEY, 0, KEY_QUERY_VALUE, &hKey) != ERROR_SUCCESS){
		return FALSE;
	}

	/* bSKeyMD5 */
	dwType   = REG_DWORD;
	dwLength = sizeof(dwData);
	if (RegQueryValueEx(hKey, SKEYMD5_VALNAME, NULL, &dwType, (LPBYTE)&dwData, &dwLength) == ERROR_SUCCESS)
		bSKeyMD5 = (BOOL)dwData;

	RegCloseKey(hKey);

	return TRUE;
}

BOOL SaveRegistory()
{
	HKEY  hKey;
	DWORD dwType;
	DWORD dwLength;
	DWORD dwData;

	if (RegCreateKeyEx(HKEY_CURRENT_USER, OTPCE_SUBKEY, 0, TEXT(""), REG_OPTION_NON_VOLATILE,
						KEY_WRITE, NULL, &hKey, &dwType) != ERROR_SUCCESS){
		return FALSE;
	}

	/* bSKeyMD5 */
	dwType   = REG_DWORD;
	dwLength = sizeof(dwData);
	dwData   = (DWORD)bSKeyMD5;
	RegSetValueEx(hKey, SKEYMD5_VALNAME, 0, dwType, (LPBYTE)&dwData, dwLength);

	RegCloseKey(hKey);

	return TRUE;
}


int CheckOption(LPTSTR lpCL)
{
	int Option;

	Option = 0;

	while (lpCL && *lpCL){
		if (!wcsnicmp(lpCL, DIALOG_MODE_OPT, wcslen(DIALOG_MODE_OPT)))
			Option |= DIALOG_MODE;
		lpCL++;
	}

	return Option;
}

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance,
                    LPTSTR lpCmdLine, int nCmdShow )
{
    HWND	hWnd;
    MSG		msg;
	int		Option;

    hInst = hInstance;      // Save handle to create command bar

	InitOtpCE();

	hWnd = CreateWindow(OTPCE_NAME, OTPCE_NAME, 0, 120, 240, 0, 0,
                        NULL, NULL, hInst, NULL);

    ShowWindow(hWnd, SW_HIDE);
    UpdateWindow(hWnd);

	Option = CheckOption(lpCmdLine);
	if (Option){
		PostMessage(hWnd, WM_OTPCE_OPT, 0, Option);
	}

    while (GetMessage(&msg, NULL, 0, 0) != FALSE){
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}

    return(msg.wParam);
}

LRESULT CALLBACK WndProc(HWND hWnd, UINT message,
                          WPARAM uParam, LPARAM lParam )
{
    switch (message){
    case WM_CREATE:
		TrayAdd(hWnd);
        break;

    case WM_HIBERNATE:
		MessageBox(NULL, TEXT("MEMORY IS LOW"), szAppName, MB_OK);
		break;

	case WM_OTPCE_NOTIFY:
		if (lParam == WM_LBUTTONDOWN){
			if (GetAsyncKeyState(VK_MENU)){
				HMENU hMenu;
				UINT  uMenuID;

				SetForegroundWindow(hWnd);
				hMenu = CreatePopupMenu();
				AppendMenu(hMenu, MF_STRING | (bSKeyMD5 ? MF_CHECKED : 0),
						   IDM_SKEYMD5, SKEYMD5_MENU);
				AppendMenu(hMenu, MF_STRING, IDM_EXIT, EXIT_MENU);
				uMenuID = TrackPopupMenu(hMenu, TPM_BOTTOMALIGN | TPM_RETURNCMD,
							GetSystemMetrics(SM_CXSCREEN) - MENU_X_MARGIN,
							GetSystemMetrics(SM_CYSCREEN) - MENU_Y_MARGIN,
							0, hWnd, NULL);
				DestroyMenu(hMenu);

				switch(uMenuID){
				case IDM_SKEYMD5:
					bSKeyMD5 = bSKeyMD5 ? FALSE : TRUE;
					SaveRegistory();
					break;
				case IDM_EXIT:
					PostMessage(hWnd, WM_CLOSE, 0, 0);
					break;
				}

			} else {
				CalcOTP(hWnd);
			}
		}
		break;

	case WM_PASSPHRASE:
	{
		int   res;

		res = DialogBox(hInst, MAKEINTRESOURCE(IDD_PASSPHRASE),
						hWnd, PassphraseDlgProc);

		if (res){
			TCHAR  pass_tmp[ONETIME_KEY_MAX_LEN];
			HLOCAL hClipMem;

			mbstowcs(pass_tmp, pass, ONETIME_KEY_MAX_LEN);
			hClipMem = LocalAlloc(LPTR, sizeof(TCHAR) * (wcslen(pass_tmp)+3));
			wcscpy((LPTSTR)hClipMem, pass_tmp);
			wcscat((LPTSTR)hClipMem, CRLF);

			if (OpenClipboard(hWnd)){
				EmptyClipboard();
				SetClipboardData(CF_UNICODETEXT, hClipMem);
				CloseClipboard();
			} else {
				LocalFree(hClipMem);
			}
		}
	}

		break;

	case WM_OTPCE_OPT:
		if (lParam & DIALOG_MODE){
			TrayDelete(hWnd);
			ShowWindow(hWnd, SW_SHOW);
			DialogBox(hInst, MAKEINTRESOURCE(IDD_DIALOG),
					  hWnd, DialogDlgProc);
			DestroyIcon(hOTPIcon);
			DestroyWindow(hWnd);
		}
		break;

    case WM_CLOSE:
		if (IDNO ==
			MessageBox(hWnd, TEXT("Do you want to exit?"),
					   szAppName, MB_YESNO|MB_DEFBUTTON2))
			break;

		TrayDelete(hWnd);
		DestroyIcon(hOTPIcon);
        DestroyWindow(hWnd);
        break;

    case WM_DESTROY:
        PostQuitMessage(0);
        break;

    default:
        return DefWindowProc(hWnd, message, uParam, lParam);
    }

    return (0);
}

int CalcOTP(HWND hWnd)
{
	if (OpenClipboard(hWnd)){
		HLOCAL hClipMem = GetClipboardData(CF_UNICODETEXT);
		if (hClipMem){
			LPTSTR tbuf = (LPTSTR)hClipMem;
			DWORD  cbsize = LocalSize(hClipMem) ;

			if (cbsize && FindSeq(tbuf)){
				PostMessage(hWnd, WM_PASSPHRASE, 0, 0);
			} else {
				goto error;
			}
		} else {
			goto error;
		}
		CloseClipboard();
		return TRUE;
	}
	return FALSE;

error:
	CloseClipboard();
	return FALSE;
}

BOOL CALLBACK PassphraseDlgProc(HWND hWnd, UINT uMessage,
				WPARAM wParam, LPARAM lParam)
{
    switch(uMessage) {
    case WM_INITDIALOG:
	{
		TCHAR Title[SEED_MAX_LEN + 10];
		TCHAR strtmp[BUF_SIZE];

		SetForegroundWindow(hWnd);

		mbstowcs(strtmp, seed, BUF_SIZE);
		wsprintf(Title, TEXT("%d %s"), seq, strtmp);
		SetWindowText(hWnd, Title);		

		SetWindowText(GetDlgItem(hWnd, IDC_FUNC), FuncName);
	}
		break;

	case WM_COMMAND:
		switch(wParam){
		case IDOK:
		{
			TCHAR passphrase_tmp[PASS_PHRASE_MAX_LEN];
			char  passphrase[PASS_PHRASE_MAX_LEN];
			char  key[8];
			int   i;

			GetDlgItemText(hWnd, IDC_PASS, passphrase_tmp, PASS_PHRASE_MAX_LEN);
			wcstombs(passphrase, passphrase_tmp, PASS_PHRASE_MAX_LEN);

			if (KeyCrunchFunc(key, seed, passphrase) == -1){
				EndDialog(hWnd, FALSE);
				break;
			}

			for (i = 0; i < seq; i++) IterateFunc(key);
			OTP_btoe(pass, key);

			EndDialog(hWnd, TRUE);
		}
			break;

		case IDCANCEL:
			EndDialog(hWnd, FALSE);
			break;

		default:
			return FALSE;
		}
		break;

	default:
		return FALSE;
	}

	return TRUE;
}																	   	

int FindSeq(LPTSTR buffer)
{
	LPTSTR p, ps;
	TCHAR  seed_tmp[SEED_MAX_LEN];
	int    i, slen;

	if ((p = ScanBuf(buffer, OTPMD4)) != NULL){
		hash_func = MD4;
	} else if ((p = ScanBuf(buffer, OTPMD5))     != NULL){
		hash_func = MD5;
	} else if ((p = ScanBuf(buffer, OTPSHA1))    != NULL){
		hash_func = SHA1;
	} else if ((p = ScanBuf(buffer, SKEY))       != NULL){
		hash_func = bSKeyMD5 ? MD5 : MD4;
	} else if ((p = ScanBuf(buffer, BSDOS_SKEYMD4)) != NULL){
		hash_func = MD4;
	} else if ((p = ScanBuf(buffer, BSDOS_SKEYMD5)) != NULL){
		hash_func = MD5;
	} else {
		return FALSE;
	}

	SetFuncName();

	p++;
	seq = _wtoi(p);

	i = 0;
	ps = NULL;
	while (p && *p){
		if (!iswalnum(*p)){
			i++;
			if (i == 2) break;
		} else if (i == 1 && ps == NULL){
			ps = p;
		}
		p++;
	}
	if (i == 1 && ps) i++;

	if (i == 2){
		slen = p - ps;
		wcsncpy(seed_tmp, ps, slen);
		seed_tmp[slen] = '\0';
		wcstombs(seed, seed_tmp, SEED_MAX_LEN);
	} else {
		return FALSE;
	}

	return TRUE;
}

LPTSTR ScanBuf(LPTSTR p, LPTSTR key)
{
	TCHAR *s, *e;
	int  klen;

	klen = wcslen(key);

	for (s = p, e = p + wcslen(p) - klen; s < e; s++){
		if (!wcsncmp(s, key, klen))
			return (s + klen);
	}

	return NULL;
}

int SetFuncName()
{
	switch(hash_func){
	case MD4:
		KeyCrunchFunc = (void *)_otp_MD4_keycrunch;
		IterateFunc   = (void *)_otp_MD4_iterate;
		wcscpy(FuncName, FUNCNAME_MD4);
		break;
	case MD5:
		KeyCrunchFunc = (void *)_otp_MD5_keycrunch;
		IterateFunc   = (void *)_otp_MD5_iterate;
		wcscpy(FuncName, FUNCNAME_MD5);
		break;
	case SHA1:
		KeyCrunchFunc = (void *)_otp_SHA1_keycrunch;
		IterateFunc   = (void *)_otp_SHA1_iterate;
		wcscpy(FuncName, FUNCNAME_SHA1);
		break;

	default:
		return FALSE;
	}

	return TRUE;
}

BOOL CALLBACK DialogDlgProc(HWND hWnd, UINT uMessage,
				WPARAM wParam, LPARAM lParam)
{
    switch(uMessage) {
    case WM_INITDIALOG:
		SetForegroundWindow(hWnd);
	{
		HWND hWndHash = GetDlgItem(hWnd, IDC_HASHFUNC);
		SendMessage(hWndHash, CB_ADDSTRING, 0, (LPARAM)FUNCNAME_MD4);
		SendMessage(hWndHash, CB_ADDSTRING, 0, (LPARAM)FUNCNAME_MD5);
		SendMessage(hWndHash, CB_ADDSTRING, 0, (LPARAM)FUNCNAME_SHA1);
		SendMessage(hWndHash, CB_SELECTSTRING, (WPARAM)-1, (LPARAM)FUNCNAME_MD5);
	}
		break;

	case WM_COMMAND:
		switch(wParam){
		case IDOK:
		{
			HWND	hWndCtrl;
			char	passphrase[PASS_PHRASE_MAX_LEN];
			char	key[8];
			int		i;

			hWndCtrl= GetDlgItem(hWnd, IDC_HASHFUNC);
 			i = SendMessage(hWndCtrl, CB_GETCURSEL, 0, 0L);
			switch(i){
			case 0:
				hash_func = MD4;
				break;
			case 2:
				hash_func = SHA1;
				break;
			case 1:
			default:
				hash_func = MD5;
				break;
			}

			SetFuncName();

			seq = GetDlgItemInt(hWnd, IDC_SEQ, NULL, FALSE);

			{
				TCHAR seed_tmp[SEED_MAX_LEN];
				GetDlgItemText(hWnd, IDC_SEED, seed_tmp, SEED_MAX_LEN);
				wcstombs(seed, seed_tmp, SEED_MAX_LEN);
			}

			{
				TCHAR passphrase_tmp[PASS_PHRASE_MAX_LEN];
				GetDlgItemText(hWnd, IDC_PASS, passphrase_tmp, PASS_PHRASE_MAX_LEN);
				wcstombs(passphrase, passphrase_tmp, PASS_PHRASE_MAX_LEN);
			}

			if (KeyCrunchFunc(key, seed, passphrase) == -1){
				EndDialog(hWnd, FALSE);
				break;
			}

			for (i = 0; i < seq; i++) IterateFunc(key);
			OTP_btoe(pass, key);

			{
				TCHAR pass_tmp[ONETIME_KEY_MAX_LEN];
				mbstowcs(pass_tmp, pass, ONETIME_KEY_MAX_LEN);
				SetDlgItemText(hWnd, IDC_OTK, pass_tmp);
			}

			{
				TCHAR tbuf[16];
				wsprintf(tbuf, TEXT("%d"), --seq);
				SetDlgItemText(hWnd, IDC_SEQ, tbuf);
			}
		}
			break;

		case IDCANCEL:
			EndDialog(hWnd, FALSE);
			break;

		default:
			return FALSE;
		}
		break;

	default:
		return FALSE;
	}

	return TRUE;
}																	   	
